#ifndef _ZRVL_VARIANT_H_
#define _ZRVL_VARIANT_H_
//--------------------------------------------------------------
// 
//--------------------------------------------------------------

#define ZRVL_DEBUG_MODE

// TODO: if TVP Plugin mode
#include "TPException.h"


// ϐ̌^
enum tZrvlVariantType {
	vtZrvl_void = 0,
	vtZrvl_null,
	vtZrvl_Integer,
	vtZrvl_Real,
	vtZrvl_String,
	vtZrvl_Octet,
	vtZrvl_Object,
	vtZrvl_Class,
	vtZrvl_Function,
	vtZrvl_Property,
	vtZrvl_Array,
	vtZrvl_Dictionary,
	vtZrvl_undefined,
	vtZrvl_Dummy_Last
};


// is null?
bool ZrvlIsNullObject(const tTJSVariant * obj);
bool ZrvlIsNullObject(const tTJSVariantClosure * obj);

// instanceof Zq
bool ZrvlIsObjectInstanceOf(iTJSDispatch2 * obj, const tjs_char * name);
bool ZrvlIsObjectInstanceOf(const tTJSVariantClosure * obj, const tjs_char * name);
bool ZrvlIsObjectInstanceOf(const tTJSVariant * var, const tjs_char * name);

// ^
const ttstr & ZrvlGetVariantTypeName(const tTJSVariant * variant);
const ttstr & ZrvlGetVariantTypeName(const tTJSVariant & variant);
const ttstr & ZrvlGetVariantTypeName(tTJSVariantType type);

// ϐ̌^
tZrvlVariantType ZrvlGetVariantType(const tTJSVariant * target);

// NXCX^Xizj
tjs_error ZrvlGetClassInstanceInfoArray(tTJSVariant * target, /*result*/tTJSVariant * arrayobj, const tjs_char * callerName);

// NXCX^X
tjs_error ZrvlGetClassInstanceInfo(tTJSVariant * target, /*result*/tTJSVariant * stringobj, const tjs_char * callerName);



//---------------------------------------------------------------------------
// 
/*final*/ class ZrvlParameterIterator
{
protected:
	tTJSVariant ** Param;
	tjs_int NumParams;
	tjs_int Index;
	tjs_int NumRequired;
	const tjs_char * Caller;

public:
	ZrvlParameterIterator(tTJSVariant ** param, tjs_int numparams, tjs_int numrequired, const tjs_char * caller = TJSEM_ANON_FUNC)
	{
		Param = param;
		NumParams = numparams;
		NumRequired = numrequired;
		Caller = caller;
		Reset();
#ifdef ZRVL_DEBUG_MODE
	ttstr message = TVPFormatMessage(TJS_W("Constructor: num=%1 req=%2"), ttstr(NumParams), ttstr(NumRequired));
	TVPAddLog(message);
#endif
	}

public:
	bool CurrentIndex()
	{
		return Index;
	}

	bool IsSufficient(void)
	{
		if (NumRequired >= 0) return (NumParams >= NumRequired);
		return true;
	}

	void Reset(void)
	{
		Index = 0;
	}

	void Next(void)
	{
		Index++;
	}

	bool HasNext(void)
	{
		return (Index < (NumParams + 1));
	}

	// reject current value if void
	bool RejectVoid(void)
	{
		bool matched = IsValid();
		if (!matched) TJSE_THROW(TJSEM_E_ARG_MUST_NOT_VOID, Caller, (ttstr)(Index + 1));
		return true;
	}

	// current value is ...
	bool IsVoid(void) { return (Param[Index]->Type() == tvtVoid); }
	bool IsString(void) { return (Param[Index]->Type() == tvtString); }
	bool IsOctet(void) { return (Param[Index]->Type() == tvtOctet); }
	bool IsObject(void) { return (Param[Index]->Type() == tvtObject); }
	bool IsClass(void) { return (ZrvlGetVariantType(Param[Index]) == vtZrvl_Class); }
	bool IsFunction(void) { return (ZrvlGetVariantType(Param[Index]) == vtZrvl_Function); }
	bool IsProperty(void) { return (ZrvlGetVariantType(Param[Index]) == vtZrvl_Property); }
	bool IsNull(void) { return (IsObject() && ZrvlIsNullObject(Param[Index])); }
	bool IsInstanceOf(const tjs_char * className) { return (IsObject() && ZrvlIsObjectInstanceOf(Param[Index], className)); }

	bool IsValid(void) {
		bool matched = (NumParams > Index) && Param[Index] && (Param[Index]->Type() != tvtVoid);
#ifdef ZRVL_DEBUG_MODE
	ttstr message = TVPFormatMessage(TJS_W("IsValid: num=%1 index=%2"), ttstr(NumParams), ttstr(Index));
	TVPAddLog(message);
	message = matched ? TJS_W("IsValid: matched") : TJS_W("IsValid: unmatched");
	TVPAddLog(message);
#endif
		return (matched);
	}

	bool IsStringOrThrow(void)
	{
		if (!IsString()) TJSE_THROW(TJSEM_E_ARG_MUST_STR, Caller, (ttstr)(Index + 1));
		return true;
	}
	bool IsOctetOrThrow(void)
	{
		if (!IsOctet()) TJSE_THROW(TJSEM_E_ARG_MUST_OCT, Caller, (ttstr)(Index + 1));
		return true;
	}
	bool IsObjectOrThrow(void)
	{
		if (!IsObject()) TJSE_THROW(TJSEM_E_ARG_MUST_OBJ, Caller, (ttstr)(Index + 1));
		return true;
	}
	bool IsClassOrThrow(void)
	{
		if (!IsClass()) TJSE_THROW(TJSEM_E_ARG_MUST_CLASS, Caller, (ttstr)(Index + 1));
		return true;
	}
	bool IsFunctionOrThrow(void)
	{
		if (!IsFunction()) TJSE_THROW(TJSEM_E_ARG_MUST_FUNCTION, Caller, (ttstr)(Index + 1));
		return true;
	}
	bool IsPropertyOrThrow(void)
	{
		if (!IsProperty()) TJSE_THROW(TJSEM_E_ARG_MUST_PROPERTY, Caller, (ttstr)(Index + 1));
		return true;
	}
	bool IsNotNullOrThrow(void)
	{
		if (IsNull()) TJSE_THROW(TJSEM_E_ARG_MUST_NOT_NULL, Caller, (ttstr)(Index + 1));
		return true;
	}
	bool IsNotVoidOrThrow(void)
	{
		if (IsVoid()) TJSE_THROW(TJSEM_E_ARG_MUST_NOT_VOID, Caller, (ttstr)(Index + 1));
		return true;
	}
	bool IsInstanceOfOrThrow(const tjs_char * className)
	{
		if (!IsInstanceOf(className)) TJSE_THROW(TJSEM_E_ARG_MUST_INSTANCE_OF, Caller, (ttstr)(Index + 1), className);
		return true;
	}

	tjs_int32 AsInteger(void) { return (tTVInteger)*Param[Index]; }
	tjs_uint32 AsUnsignedInteger(void) { return (tTVInteger)*Param[Index]; }
	tjs_int64 AsInteger64(void) { return (tTVInteger)*Param[Index]; }
	tjs_real AsReal(void) { return (tTVReal)*Param[Index]; }
	ttstr AsString(void) { return *Param[Index]; }
	iTJSDispatch2 * AsObject(void) { return Param[Index]->AsObject(); }
	iTJSDispatch2 * AsObjectNoAddRef(void) { return Param[Index]->AsObjectNoAddRef(); }
	iTJSDispatch2 * AsObjectThis(void) { return Param[Index]->AsObjectThis(); }
	iTJSDispatch2 * AsObjectThisNoAddRef(void) { return Param[Index]->AsObjectThisNoAddRef(); }
	tTJSVariantClosure * AsObjectClosure(void) { return &(Param[Index]->AsObjectClosure()); }
	tTJSVariantClosure * AsObjectClosureNoAddRef(void) { return &(Param[Index]->AsObjectClosureNoAddRef()); }
	tTJSVariantOctet * AsVariantOctet(void) { return Param[Index]->AsOctet(); }
	tTJSVariantOctet * AsVariantOctetNoAddRef(void) { return Param[Index]->AsOctetNoAddRef(); }
	tTJSVariantString * AsVariantString(void) { return Param[Index]->AsString(); }
	tTJSVariantString * AsVariantStringNoAddRef(void) { return Param[Index]->AsStringNoAddRef(); }
	tTJSVariant * AsVariant(void) { return Param[Index]; }

	// default
	tjs_int32 AsInteger(tjs_int32 d) { return (IsValid() ? AsInteger() : d); }
	tjs_uint32 AsUnsignedInteger(tjs_uint32 d) { return (IsValid() ? AsUnsignedInteger() : d); }
	tjs_int64 AsInteger64(tjs_int64 d) { return (IsValid() ? AsInteger64() : d); }
	tjs_real AsReal(tjs_real d) { return (IsValid() ? AsReal() : d); }
	ttstr AsString(ttstr & d) { return (IsValid() ? AsString() : d); }
	iTJSDispatch2 * AsObject(iTJSDispatch2 * d) { return (IsValid() ? AsObject() : d); }
	iTJSDispatch2 * AsObjectNoAddRef(iTJSDispatch2 * d) { return (IsValid() ? AsObjectNoAddRef() : d); }
	iTJSDispatch2 * AsObjectThis(iTJSDispatch2 * d) { return (IsValid() ? AsObjectThis() : d); }
	iTJSDispatch2 * AsObjectThisNoAddRef(iTJSDispatch2 * d) { return (IsValid() ? AsObjectThisNoAddRef() : d); }
	tTJSVariantClosure * AsObjectClosure(tTJSVariantClosure * d) { return (IsValid() ? AsObjectClosure() : d); }
	tTJSVariantClosure * AsObjectClosureNoAddRef(tTJSVariantClosure * d) { return (IsValid() ? AsObjectClosureNoAddRef() : d); }
	tTJSVariantOctet * AsVariantOctet(tTJSVariantOctet * d) { return (IsValid() ? AsVariantOctet() : d); }
	tTJSVariantOctet * AsVariantOctetNoAddRef(tTJSVariantOctet * d) { return (IsValid() ? AsVariantOctetNoAddRef() : d); }
	tTJSVariantString * AsVariantString(tTJSVariantString * d) { return (IsValid() ? AsVariantString() : d); }
	tTJSVariantString * AsVariantStringNoAddRef(tTJSVariantString * d) { return (IsValid() ? AsVariantStringNoAddRef() : d); }
	tTJSVariant * AsVariant(tTJSVariant * d) { return (IsValid() ? AsVariant() : d); }

	ttstr AsString(const tjs_char * d) { return (IsValid() ? AsString() : ttstr(d)); }

};
//---------------------------------------------------------------------------


//--------------------------------------------------------------
#endif  _ZRVL_VARIANT_H_
