#ifndef _TVP_PLUGIN_DATEUTIL_INTF_H_
#define _TVP_PLUGIN_DATEUTIL_INTF_H_

#include "TVPPlugIn.h"
#include "JulianDay.h"


struct TJSETimestampCache
{
	tjs_int y, m, d;	// N
	tjs_int hh, mi, ss;	// b
	tjs_real realsec;	// bij
	tjs_int day;	// day of weekijj
	tjs_int doy;	// days of yeariNʓj
	bool leap;	// leap year
	bool enabled;	// is cache enabled?
};

// rp
typedef tjs_uint32 TJSEDateComparisonFlags_t;
#define TJSE_DATE_CMP_FLAG_NONE    (0UL)
#define TJSE_DATE_CMP_FLAG_YEAR    (1UL << 0)
#define TJSE_DATE_CMP_FLAG_MONTH   (1UL << 1)
#define TJSE_DATE_CMP_FLAG_DATE    (1UL << 2)
#define TJSE_DATE_CMP_FLAG_HOURS   (1UL << 3)
#define TJSE_DATE_CMP_FLAG_MINUTES (1UL << 4)
#define TJSE_DATE_CMP_FLAG_SECONDS (1UL << 5)
#define TJSE_DATE_CMP_FLAG_REALSEC (1UL << 6)
#define TJSE_DATE_CMP_FLAG_WEEKDAY (1UL << 7)

// W3CDTF
#define TJSE_W3CDTF_FORMAT_DATE TJS_W("%04d-%02d-%02d")
#define TJSE_W3CDTF_FORMAT_TIME TJS_W("%02d:%02d:%06.3f")
#define TJSE_W3CDTF_FORMAT_TZ   TJS_W("%+03d:%02d")
#define TJSE_W3CDTF_FORMAT_DT   (TJSE_W3CDTF_FORMAT_DATE TJS_W("T") TJSE_W3CDTF_FORMAT_TIME)
#define TJSE_W3CDTF_FORMAT_FULL (TJSE_W3CDTF_FORMAT_DT TJSE_W3CDTF_FORMAT_TZ)


// lCeBuCX^X
class TJSE_NI_CLASS_NAME(Timestamp);

//---------------------------------------------------------------------------
// lCeBuCX^X
//---------------------------------------------------------------------------
TJSE_NATIVE_INSTANCE_DECL_BEGIN(Timestamp)
protected:
	// current time in Julian day (UTC)
	// ̒lSĂ̊ƂȂB
	tjs_real JD;	// Ə
	tjs_real JDTime;	// ̂

	// current timezone
	const jd_TimezoneLookupTable *Timezone;

	// local time, Gregorian
	TJSETimestampCache Cache;

	// enabled?
	bool Enabled;

	// Ă񂾐l
	const tjs_char * NCMName;

public:
	// C++ RXgN^
	TJSE_NI_CONSTRUCTOR(Timestamp);

	// C++fXgN^
	TJSE_NI_DESTRUCTOR(Timestamp);

	// TJS RXgN^
	TJSE_TJS_CONSTRUCTOR_DECL
	{
		return S_OK;
	}

	// TJS fXgN^
	void TJS_INTF_METHOD Invalidate()
	{
	}

protected:
	// cache control
	virtual void FireChanged();
	virtual void CreateCache();

	inline bool IsCacheEnabled() const
	{
		return Cache.enabled;
	}

public:
	// O΍
	inline void SetNCM(const tjs_char * name)
	{
		NCMName = name;
	}

	// O΍
	inline void ClearNCM()
	{
		NCMName = NULL;
	}

	inline const TJSETimestampCache & CacheRef() const
	{
		return Cache;
	}

	inline tjs_real GetJulianDay() const
	{
		return JD;
	}

	inline tjs_real GetJulianDayTime() const
	{
		return JDTime;
	}

	inline bool IsEnabled() const
	{
		return Enabled;
	}

protected:
	inline bool Enable()
	{
		Enabled = true;
	}

	inline bool Disable()
	{
		Enabled = false;
	}

public:
	tjs_int GetTimezone() const;	// in seconds
	bool GetDST() const;
	const tjs_char * GetTimezoneName() const;	// abbr (lower case)
	const tjs_char * GetTimezoneFullName() const;	// full

	void SetTimezone(const tjs_char *name, bool dst);	// abbr, full

	void AddDateTime(tjs_int y, tjs_int m, tjs_int d, tjs_int hh, tjs_int mi, tjs_real ss);
	tjs_int CompareTo(TJSE_NI_CLASS_NAME(Timestamp) * target, TJSEDateComparisonFlags_t flags) const;

	void Assign(tjs_int tz, tjs_int y, tjs_int m, tjs_int d, tjs_int hh, tjs_int mi, tjs_real ss);
	void Assign(tjs_int y, tjs_int m, tjs_int d, tjs_int hh, tjs_int mi, tjs_real ss);
	void Assign(const TJSE_NI_CLASS_NAME(Timestamp) * target);	// copy
//	void Assign(iTJSDispatch2 * dateobj);	// TJS Date
	void Assign(tTJSVariantClosure * datevclo);	// TJS Date
	void Assign();	// now
	void Assign(tjs_real jd);
	void Assign(tjs_real jd, tjs_real jdt);	// core

//	tjs_int GetEpoch();	// in seconds
	tjs_int GetLastDayOfMonth() const;

	// to string
	void GetDateString(ttstr &result) const;
	void GetTimeString(ttstr &result) const;
	void GetString(ttstr &result) const;

	void GetLocaleString(ttstr &result, const tjs_char * tzname) const;
	void GetLocaleDateString(ttstr &result, const tjs_char * tzname) const;
	void GetLocaleTimeString(ttstr &result, const tjs_char * tzname) const;

TJSE_NATIVE_INSTANCE_DECL_END


#endif _TVP_PLUGIN_DATEUTIL_INTF_H_

