#include "JulianDay.h"
#include <math.h>


//----------------------------------------------------------------------
/*
jd_GetJulianDay() and jd_GetYMDHMS() are based on the following functions.
	GregorianToSdn()
	SdnToGregorian()
see php-5.2.0.tar.gz:/php-5.2.0/ext/calendar/gregor.c
*/
/* $selId: gregor.c,v 2.0 1995/10/24 01:13:06 lees Exp $
 * Copyright 1993-1995, Scott E. Lee, all rights reserved.
 * Permission granted to use, copy, modify, distribute and sell so long as
 * the above copyright and this permission statement are retained in all
 * copies.  THERE IS NO WARRANTY - USE AT YOUR OWN RISK.
 */
#define PHP5_GREGOR_SDN_OFFSET  32045
#define PHP5_DAYS_PER_5_MONTHS  153
#define PHP5_DAYS_PER_4_YEARS   1461
#define PHP5_DAYS_PER_400_YEARS 146097


//----------------------------------------------------------------------
#ifdef JD_REQUIRE_TJS_STRICMP
static tjs_int TJS_stricmp(const tjs_char *s1, const tjs_char *s2)
{
	// we only support basic alphabets
	// fixme: complete alphabets support

	for(;;)
	{
		tjs_char c1 = *s1, c2 = *s2;
		if(c1 >= TJS_W('a') && c1 <= TJS_W('z')) c1 += TJS_W('Z')-TJS_W('z');
		if(c2 >= TJS_W('a') && c2 <= TJS_W('z')) c2 += TJS_W('Z')-TJS_W('z');
		if(c1 == TJS_W('\0')) return (c2 == TJS_W('\0')) ? 0 : -1;
		if(c2 == TJS_W('\0')) return (c1 == TJS_W('\0')) ? 0 : 1;
		if(c1 < c2) return -1;
		if(c1 > c2) return 1;
		s1++;
		s2++;
	}
}
#endif
//----------------------------------------------------------------------
/*
Date(locale) -> JD(UTC)
m : 1..12
tz in sec
*/
double jd_GetJulianDay(tjs_int tz, tjs_int y, tjs_int m, tjs_int d, 
	tjs_int hh, tjs_int mi, tjs_real ss, tjs_real &time)
{
	tjs_int year, month;
	const tjs_real errorcode = -1.0;

	/* check for invalid dates */
	if (y == 0 || y < -4714 || m <= 0 || m > 12 || d <= 0 || d > 31) return (errorcode);

	/* check for dates before SDN 1 (Nov 25, 4714 B.C.) */
	if ((y == -4714) && (m < 11 || (m == 11 && d < 25))) return (errorcode);

	/* Make year always a positive number. */
	year = y + 4800;
	if (y < 0) year++;

	/* Adjust the start of the year. */
	if (m > 2)
	{
		month = m - 3;
	}
	else
	{
		month = m + 9;
		year--;
	}

	tjs_real jd = ((year / 100) * PHP5_DAYS_PER_400_YEARS) / 4
			+ ((year % 100) * PHP5_DAYS_PER_4_YEARS) / 4
			+ (month * PHP5_DAYS_PER_5_MONTHS + 2) / 5
			+ d - PHP5_GREGOR_SDN_OFFSET;	// 0ɂȂB

	tjs_real t = ss + (tjs_real)mi * 60.0 + (tjs_real)hh * 3600.0 - tz;	// sec
	t = t / 86400.0 - 0.5;
	time = t - ::floor(t);
	return (jd + t);
}
//----------------------------------------------------------------------
void jd_GetYMDHMS(tjs_real jd, tjs_int tz, 
	tjs_int &y, tjs_int &m, tjs_int &d, tjs_int &hh, tjs_int &mi, tjs_real &ss)
{
	tjs_real time;	// dummy
	jd_GetYMDHMS(jd, time, tz, y, m, d, hh, mi, ss);
}
//----------------------------------------------------------------------
void jd_GetYMDHMS(tjs_real jd, tjs_real time, tjs_int tz, 
	tjs_int &y, tjs_int &m, tjs_int &d, tjs_int &hh, tjs_int &mi, tjs_real &ss)
{
	const tjs_int errorcode = 0;
	const tjs_real tztime = (tjs_real)tz / 86400.0;
	jd += tztime + 0.5;

	tjs_int sdn = jd;
	tjs_int century;
	tjs_int year, month, day;
	tjs_int temp;
	tjs_int dayOfYear;

	if (sdn <= 0)
	{
		y = m = d = errorcode;
		return;
	}
	temp = (sdn + PHP5_GREGOR_SDN_OFFSET) * 4 - 1;

	if (temp < 0)
	{
		y = m = d = errorcode;
		return;
	}

	/* Calculate the century (year/100). */
	century = temp / PHP5_DAYS_PER_400_YEARS;

	/* Calculate the year and day of year (1 <= dayOfYear <= 366). */
	temp = ((temp % PHP5_DAYS_PER_400_YEARS) / 4) * 4 + 3;
	year = (century * 100) + (temp / PHP5_DAYS_PER_4_YEARS);
	dayOfYear = (temp % PHP5_DAYS_PER_4_YEARS) / 4 + 1;

	/* Calculate the month and day of month. */
	temp = dayOfYear * 5 - 3;
	month = temp / PHP5_DAYS_PER_5_MONTHS;
	day = (temp % PHP5_DAYS_PER_5_MONTHS) / 5 + 1;

	/* Convert to the normal beginning of the year. */
	if (month < 10)
	{
		month += 3;
	}
	else
	{
		year += 1;
		month -= 9;
	}

	/* Adjust to the B.C./A.D. type numbering. */
	year -= 4800;
	if (year <= 0) year--;

	y = year;
	m = month;
	d = day;

	tjs_real wt = time + tztime + 0.5;
	tjs_real tm = 86400.0 * (wt - ::floor(wt));	// bɊZ

	hh = tjs_int(tm / 3600.0);
	mi = tjs_int((tm - 3600.0 * (tjs_real)hh) / 60.0);
	ss = tm - 3600.0 * (tjs_real)hh - 60.0 * (tjs_real)mi;
	return;
}
//----------------------------------------------------------------------
bool jd_IsLeapYear(tjs_int y)
{
	return (!(y % 400) || (!(y % 4) && (y % 100)));
}



//----------------------------------------------------------------------
/* The timezone table. */
// php-5.2.0/ext/date/lib/timezonemap.h
// php-5.2.0/ext/date/lib/fallbackmap.h
// php-5.2.0/ext/date/lib/parse_date.c
const static jd_TimezoneLookupTable timelib_timezone_lookup[] = {
#include "php5_timezonemap.h"
	{ NULL, 0, 0, NULL },
};

const static jd_TimezoneLookupTable timelib_timezone_fallbackmap[] = {
#include "php5_fallbackmap.h"
	{ NULL, 0, 0, NULL },
};

const static jd_TimezoneLookupTable timelib_timezone_utc[] = {
	{ TJS_W("utc"), 0, 0, TJS_W("UTC") },
};
//----------------------------------------------------------------------
const jd_TimezoneLookupTable* jd_SearchTimezone(const tjs_char *word, 
	tjs_int gmtoffset, tjs_int isdst, bool fullname)
{
	bool first_found = false;
	const jd_TimezoneLookupTable *tp, *first_found_elem = NULL;
	const jd_TimezoneLookupTable *fmp;

	if (TJS_stricmp(TJS_W("utc"), word) == 0 || TJS_stricmp(TJS_W("gmt"), word) == 0) {
		return timelib_timezone_utc;
	}

	for (tp = timelib_timezone_lookup; tp->name; tp++) {
		const tjs_char * abbr = tp->name;
		const tjs_char * full = tp->full_tz_name;
		if ((!fullname && abbr && TJS_stricmp(word, abbr) == 0) || 
			( fullname && full && TJS_stricmp(word, full) == 0) )
		{
			if (!first_found) {
				first_found = true;
				first_found_elem = tp;
				if (gmtoffset == -1) {
					return tp;
				}
			}
			if (tp->gmtoffset == gmtoffset) {
				return tp;
			}
		}
	}
	if (first_found) {
		return first_found_elem;
	}
	return NULL;
}
//----------------------------------------------------------------------
const jd_TimezoneLookupTable* jd_SearchTimezone(const tjs_char *word, 
	tjs_int isdst, bool fullname)
{
	bool first_found = false;
	const jd_TimezoneLookupTable *tp, *first_found_elem = NULL;

	if (TJS_stricmp(TJS_W("utc"), word) == 0 || TJS_stricmp(TJS_W("gmt"), word) == 0) {
		return timelib_timezone_utc;
	}

	for (tp = timelib_timezone_lookup; tp->name; tp++) {
		const tjs_char * abbr = tp->name;
		const tjs_char * full = tp->full_tz_name;
		if ((!fullname && abbr && TJS_stricmp(word, abbr) == 0) || 
			( fullname && full && TJS_stricmp(word, full) == 0) )
		{
			if (!first_found) {
				first_found = true;
				first_found_elem = tp;
				if (tp->type == isdst) {
					return tp;
				}
			}
		}
	}
	if (first_found) {
		return first_found_elem;
	}
	return NULL;
}
//----------------------------------------------------------------------
const jd_TimezoneLookupTable* jd_SearchTimezoneFB(const tjs_char *word, 
	tjs_int isdst, bool fullname)
{
	if (TJS_stricmp(TJS_W("utc"), word) == 0 || TJS_stricmp(TJS_W("gmt"), word) == 0) {
		return timelib_timezone_utc;
	}

	bool first_found = false;
	const jd_TimezoneLookupTable *fmp, *first_found_elem = NULL;

	for (fmp = timelib_timezone_fallbackmap; fmp->name; fmp++) {
		const tjs_char * abbr = fmp->name;
		const tjs_char * full = fmp->full_tz_name;
		if ((!fullname && abbr && TJS_stricmp(word, abbr) == 0) || 
			( fullname && full && TJS_stricmp(word, full) == 0) )
		{
			if (fmp->type == isdst) return fmp;
		}
	}
	return NULL;
}
//----------------------------------------------------------------------
const tjs_char *jd_FullZoneNameFromAbbr(const tjs_char *abbr, 
	tjs_int gmtoffset, tjs_int isdst)
{
	const jd_TimezoneLookupTable *tp;

	tp = jd_SearchTimezone(abbr, gmtoffset, isdst, false);
	if (tp) {
		return (tp->full_tz_name);
	} else {
		return NULL;
	}
}
//----------------------------------------------------------------------
