#include "SharedMemoryNI.h"


#define GET_SIZE_LOW(_s) (_s & 0xFFFFFFFFUL)


//--------------------------------------------------------------
TJSE_NI_CLASS_NAME(SharedMemory)::TJSE_NI_CONSTRUCTOR(SharedMemory)
{
	LastErrorCode = 0;
	InternalInitGet();
	InternalInitAtt();
}

//--------------------------------------------------------------
TJSE_NI_CLASS_NAME(SharedMemory)::TJSE_NI_DESTRUCTOR(SharedMemory)
{
	Free();
}

//--------------------------------------------------------------
void TJSE_NI_CLASS_NAME(SharedMemory)::InternalInitGet()
{
	hMappingObject = 0;
	MappedSize = 0;
	ObjectName = TJS_W("");
}

//--------------------------------------------------------------
void TJSE_NI_CLASS_NAME(SharedMemory)::InternalInitAtt()
{
	BaseAddress = NULL;
	AccessMode = 0;
}

//--------------------------------------------------------------
void TJSE_NI_CLASS_NAME(SharedMemory)::InternalRelease()
{
	return;
}

//--------------------------------------------------------------
tjs_error TJSE_NI_CLASS_NAME(SharedMemory)::Get(const ttstr & key, tTVInteger size)
{
	tjs_error rv = TJS_S_TRUE;	// create
	tjs_uint32 size_low = GET_SIZE_LOW(size);
	tjs_uint flag = PAGE_READWRITE;
	const size_t bufsize = 256;
	char buffer[bufsize];

	// free
	Free();

	// i[
	tjs_int narrow_len = key.GetNarrowStrLen();
	if ((narrow_len >= bufsize - 1) || (narrow_len == -1))
		return TJS_E_FAIL;
	key.ToNarrowStr(buffer, narrow_len);

	HANDLE h = CreateFileMapping((HANDLE)0xFFFFFFFF, NULL, flag, 0, size_low, buffer);
	LastErrorCode = GetLastError();
	if (h == 0) {
		rv = TJS_E_FAIL;	// failed to map
	} else if (LastErrorCode == ERROR_ALREADY_EXISTS) {
		rv = TJS_S_FALSE;	// already exists
	}

	if (TJS_SUCCEEDED(rv)) {
		ObjectName = key;
		hMappingObject = h;
		MappedSize = size_low;
	}

	return rv;
}

//--------------------------------------------------------------
// see http://support.microsoft.com/kb/830783/ja
tjs_error TJSE_NI_CLASS_NAME(SharedMemory)::Attach(tjs_uint32 access)
{
	tjs_error rv = TJS_S_OK;

	if (IsAllocated()) {
		Detach();
		LPCVOID addr = MapViewOfFile(hMappingObject, access, 0, 0, 0);
		LastErrorCode = GetLastError();
		if (!addr) {
			rv = TJS_E_FAIL;
		} else {
			BaseAddress = addr;
			AccessMode = access;
		}
	} else {
		rv = TJS_E_FAIL;
	}

	return rv;
}

//--------------------------------------------------------------
tjs_error TJSE_NI_CLASS_NAME(SharedMemory)::Detach()
{
	tjs_error rv = TJS_S_OK;

	if (IsAttached()) {
		UnmapViewOfFile(BaseAddress);
		LastErrorCode = GetLastError();
		InternalInitAtt();
	} else {
		rv = TJS_E_FAIL;
	}

	return rv;
}

//--------------------------------------------------------------
tjs_error TJSE_NI_CLASS_NAME(SharedMemory)::Free()
{
	tjs_error rv = TJS_S_OK;

	rv = Detach();
	if (IsAllocated()) {
		CloseHandle(hMappingObject);
		LastErrorCode = GetLastError();
		InternalInitGet();
	} else {
		rv = TJS_E_FAIL;
	}

	return rv;
}

//--------------------------------------------------------------
// ǂݍ
tjs_error TJSE_NI_CLASS_NAME(SharedMemory)::Peek(tTVInteger offset, tTVInteger size, tTJSVariantOctet * & data) const
{
	tjs_error rv = TJS_S_OK;
	tjs_uint32 size_low = GET_SIZE_LOW(size);
	tjs_uint32 ofs_low = GET_SIZE_LOW(offset);

	if (!IsAllocated()) return TJS_E_FAIL;
	if (ofs_low + size_low > MappedSize) return TJS_E_FAIL;

	data = TJSAllocVariantOctet((const tjs_uint8 *)BaseAddress + ofs_low, size_low);
	if (!data) rv = TJS_E_FAIL;

	return rv;
}
//--------------------------------------------------------------
// 
tjs_error TJSE_NI_CLASS_NAME(SharedMemory)::Poke(tTVInteger offset, tTVInteger size, const tTJSVariantOctet * data) const
{
	tjs_error rv = TJS_S_OK;
	tjs_uint32 size_low = GET_SIZE_LOW(size);
	tjs_uint32 ofs_low = GET_SIZE_LOW(offset);

	if (!IsAllocated()) return TJS_E_FAIL;
	if (!(AccessMode & FILE_MAP_WRITE)) return TJS_E_FAIL;
	if (ofs_low + size_low > MappedSize) return TJS_E_FAIL;
	/*ASSERT*/if (!data) return TJS_E_FAIL;

	const tjs_int srclen = data->GetLength();
	const tjs_uint8 * src = data->GetData();
	tjs_uint8 * dest = (tjs_uint8 *)BaseAddress + ofs_low;
	tjs_uint32 wlen = (srclen < size_low) ? srclen : size_low;
	while (wlen-- > 0) *dest++ = *src++;

	return rv;
}
//--------------------------------------------------------------
// ǂݍ
tjs_error TJSE_NI_CLASS_NAME(SharedMemory)::Peek32(tTVInteger offset, tjs_uint32 & data) const
{
	tjs_error rv = TJS_S_OK;
	tjs_uint32 ofs_low = GET_SIZE_LOW(offset);

	if (!IsAllocated()) return TJS_E_FAIL;
	if (ofs_low + 4 > MappedSize) return TJS_E_FAIL;

	const tjs_uint8 * p = (const tjs_uint8 *)BaseAddress + ofs_low;
	union _UINT32 { tjs_uint32 ui32; tjs_uint8 b[4]; } t;
	for (int i = 0; i < 4; i++) t.b[i] = p[i];
	data = t.ui32;

	return rv;
}
//--------------------------------------------------------------
// 
tjs_error TJSE_NI_CLASS_NAME(SharedMemory)::Poke32(tTVInteger offset, tjs_uint32 data) const
{
	tjs_error rv = TJS_S_OK;
	tjs_uint32 ofs_low = GET_SIZE_LOW(offset);

	if (!IsAllocated()) return TJS_E_FAIL;
	if (!(AccessMode & FILE_MAP_WRITE)) return TJS_E_FAIL;
	if (ofs_low + 4 > MappedSize) return TJS_E_FAIL;

	tjs_uint8 * p = (tjs_uint8 *)BaseAddress + ofs_low;
	union _UINT32 { tjs_uint32 ui32; tjs_uint8 b[4]; } t;
	t.ui32 = data;
	for (int i = 0; i < 4; i++) p[i] = t.b[i];

	return rv;
}
//--------------------------------------------------------------


