#!/usr/bin/perl#
# ネイティブクラス変換ツール
#

use strict;
use utf8;
use Encode;

use File::Basename;
use lib (dirname($0));
use TVPPlugin;

binmode(STDIN,  $ENCODE_NAME);
binmode(STDOUT, $ENCODE_NAME);

# 1: plugin mode; 0: native mode
my $PluginMode = $ARGV[0];
# Is current static ? 
my $StaticDecl = 0;
# Is current no instance class?
my $NoInstance = 0;

my $FileName = '';
my $CurrentPP = '';

my %UsingClasses = ();

my $ExtPluginDefFile = 'plugin_classes.txt';
my @ExtDefMembers = ();

my $TJSNC_PREFIX = 'tjse_NativeClassClass_';
my $TJSNI_PREFIX = 'tjse_NativeInstanceClass_';

my $TS_CONSTRUCTOR       = 'constructor';
#my $TS_FINALIZE          = 'finalize';
my $TS_EMPTY_FINALIZE    = 'emptyfinalize';
my $TS_EMPTY_CONSTRUCTOR = 'emptyconstructor';

my $TS_FUNCTION  = 'function';
my $TS_PROPERTY  = 'property';
my $TS_METHOD    = 'method';
my $TS_GETTER    = 'getter';
my $TS_SETTER    = 'setter';

my $TS_PARAMS    = 'params';
my $TS_PARAMETER = 'par';
my $TS_PARAMMUST = 'must';
my $TS_RETURN    = 'return';
my $TS_USE_CLASS = 'use-class';

my $TS_BEGIN    = 'begin';
my $TS_END      = 'end';
my $TS_FILE     = 'file';
my $TS_COMMENT  = 'comment';
my $TS_DOC      = 'doc';

my $TS_STATIC_DECL = 'static';
my $TS_NO_INSTANCE = 'noinstance';
my $TS_READONLY    = 'readonly';

my $TS_SEP = '//' . ('-' x 64);

my $TS_MEMBER_NAME = '__tp_member_name';
my $TS_CONST_MEMBER_NAME_DECL = "static const tjs_char * const ${TS_MEMBER_NAME}";
my $TS_MSG_PREFIX = $PluginMode ? 'TJSEM_' : 'ZRVLM_';

#print STDERR @ARGV . ($PluginMode ? "Plugin Mode.\n" : "Native Mode\n");

main();

sub main
{
	# 出力バッファ
	my @OUT = ();

	my @RMode = ();
	my $Class = '';
	my $Member = '';
	my $Line = 0;
	my $currentMode = '';

	while (<STDIN>) {
		chomp;
		my $a = $_;
		$Line++;

		$CurrentPP = qq(#line $Line "$FileName");

		next if ($a =~ /^\@${TS_COMMENT}/);

		if ($a =~ /^\@${TS_FILE}\s+(\S+)/) {
			$FileName = $1;
			$CurrentPP = qq(#line $Line "$FileName");
			push @OUT, "${CurrentPP}";
			next;
		}

		# end of block
		if (@RMode && ($a eq '@;')) {
			my $mode = pop @RMode;
			my $o = '';

			if ($mode eq $TS_CONSTRUCTOR) {
				$o = getConstructorDeclEnd($Class);
				$StaticDecl = 0;

			} elsif ($mode eq $TS_METHOD) {
				$o = getMethodDeclEnd($Class, $Member);
				$StaticDecl = 0;

			} elsif ($mode eq $TS_PROPERTY) {
				$o = getPropDeclEnd($Class, $Member);
				$StaticDecl = 0;

			} elsif ($mode eq $TS_GETTER) {
				$o = getGetterDeclEnd($Class, $Member);

			} elsif ($mode eq $TS_SETTER) {
				$o = getSetterDeclEnd($Class, $Member);

			} elsif ($mode eq $TS_FUNCTION) {
				$o = getFunctionDeclEnd($Member);

			} elsif ($mode eq $TS_DOC) {
				# Nothing to do

			}

			push @OUT, $o;

			next;

		# @で始まったらコマンド
		} elsif ($a =~ /^\s*\@(\S+)(.*)/) {

			my ($mode, $fullparam) = ($1, $2);
			$fullparam =~ s/^\s*//;
			my @params = split(/\s+/, $fullparam);
			my $param = $params[0];
			my $parent = @RMode ? $RMode[$#RMode] : '';

			checkNest($mode, $parent, $Line, $param);

			if ($mode eq $TS_END) {
				@RMode = ();
				push @OUT, getClassCreatorEnd($Class);
				$NoInstance = 0;
				$Class = '';
				next;

			} elsif ($mode eq "${TS_USE_CLASS}") {
				# 一行で完結するのでスタックには積まない
				push @OUT, getDeclareClassID($param);
				next;

			} elsif ($mode eq $TS_CONSTRUCTOR) {
				push @RMode, $mode;
				$StaticDecl = ($param eq $TS_STATIC_DECL);
				push @OUT, getConstructorDeclBegin($Class);
				next;

# Finalize
# finalizeメソッドはCのプリプロセッサマクロが用意されていないので、
# 通常のメソッド定義で代用してください。

			} elsif ($mode eq "${TS_EMPTY_FINALIZE};") {
				# 一行で完結するのでスタックには積まない
				push @OUT, getEmptyFinalizeDecl($Class);
				next;

			} elsif ($mode eq "${TS_EMPTY_CONSTRUCTOR};") {
				# 一行で完結するのでスタックには積まない
				push @OUT, getEmptyConstructorDecl($Class);
				next;

			} elsif ($mode eq $TS_METHOD) {
				push @RMode, $mode;
				$Member = $param;
				$StaticDecl = ($params[1] eq $TS_STATIC_DECL);
				push @OUT, getMethodDeclBegin($Class, $Member);
				next;

			} elsif ($mode eq $TS_PROPERTY) {
				# property NAME [static] [readonly TYPE VALUE]
				# TYPE = INT|REAL|OBJ|VAR|STR|PTR
				$Member = $param;
				$StaticDecl = ($params[1] eq $TS_STATIC_DECL);
				my $ro = ($params[$StaticDecl ? 2 : 1] eq $TS_READONLY);
				if ($ro) {
					my @tp = @params;
					shift @tp;
					shift @tp;
					shift @tp if ($StaticDecl);
					# 一行で完結するのでスタックには積まない
					my $type  = shift @tp;
					my $value = join(' ', @tp);
					# グローバル？
					if ($Class eq '') {
						push @ExtDefMembers, join(' ', "property $Member readonly $type $value");
						push @OUT, "// property $Member (${type}) -> $ExtPluginDefFile";
					} else {
						push @OUT, getPropDeclReadOnly($Class, $Member, $type, $value);
					}
				} else {
					push @RMode, $mode;
					push @OUT, getPropDeclBegin($Class, $Member);
				}
				next;

			} elsif ($mode eq $TS_GETTER) {
				if ($param eq 'deny;') {
					# 一行で完結するのでスタックには積まない
					push @OUT, getDenyGetterDecl($Class, $Member);
				} else {
					push @RMode, $mode;
					push @OUT, getGetterDeclBegin($Class, $Member);
				}
				next;

			} elsif ($mode eq $TS_SETTER) {
				if ($param eq 'deny;') {
					# 一行で完結するのでスタックには積まない
					push @OUT, getDenySetterDecl($Class, $Member);
				} else {
					push @RMode, $mode;
					push @OUT, getSetterDeclBegin($Class, $Member);
				}
				next;

			} elsif ($mode eq $TS_PARAMS) {
				# 一行で完結するのでスタックには積まない
				# @params NUM
				if ($parent eq $TS_SETTER) {
					push @OUT, getNumParamsCheckerForSetter($param);
				} else {
					push @OUT, getNumParamsChecker($param);
				}
				next;

			} elsif ($mode eq $TS_PARAMETER) {
				# 一行で完結するのでスタックには積まない
				# @par TYPE VARNAME [DEFAULT]
				push @OUT, getParameterGetter(split(/\s+/, $fullparam, 3));
				next;

			} elsif ($mode eq $TS_PARAMMUST) {
				# 一行で完結するのでスタックには積まない
				# @must TYPE [CLASSNAME]
				push @OUT, getParameterMust(split(/\s+/, $fullparam, 2));
				next;

			} elsif ($mode eq $TS_RETURN) {
				# 一行で完結するのでスタックには積まない
				# @return TYPE VALUE
				push @OUT, getReturnStatement(split(/\s+/, $fullparam, 2));
				next;

			} elsif ($mode eq $TS_FUNCTION) {
				$Member = $param;
				$StaticDecl = ($params[1] eq $TS_STATIC_DECL);
				push @RMode, $mode;
				push @OUT, getFunctionDeclBegin($param);
				push(@ExtDefMembers, "function $param" . ($StaticDecl ? ' static' : ''));
				next;

			} elsif ($mode eq $TS_BEGIN) {
				$Class = $param;
				shift @params;
				if ($params[0] eq $TS_NO_INSTANCE) {
					$NoInstance = 1;
					shift @params;
				} else {
					$NoInstance = 0;
				}
				push @RMode, $mode;
				push @OUT, getDefClassID($Class);
				push @OUT, getInstanceCreator($Class);
				push @OUT, getClassCreatorBegin($Class);
				push @ExtDefMembers, join(' ', "class $Class",  @params);
				next;

			} elsif ($mode eq $TS_DOC) {
				# Nothing to do
				push @RMode, $mode;
				next;

			} elsif ($currentMode ne $TS_DOC) {
				abort($Line, "Syntax error: '$a'?");

			}

		}

		$currentMode = @RMode ? $RMode[$#RMode] : '';

		next if ($currentMode eq $TS_DOC);

		push @OUT, $a;
	}

	if (@RMode) {
		abort($Line, "Nest error: unexpected EOF.");
	}

	print join("\n", @OUT, '');
	outputExtDefs();

}

sub outputExtDefs
{
	my @lines = ();

	if (open (IF, "$ExtPluginDefFile")) {
		while (<IF>) {
			chomp;
			push @lines, $_;
		}
		close(IF);
	}

	push @lines, @ExtDefMembers;
	uniqueArray(\@lines);

	open (OF, ">$ExtPluginDefFile") or die("$ExtPluginDefFile: $!");
	print OF join("\n", @lines, "");
	close(OF);
}

sub abort
{
	my ($line, $errmes) = @_;
	die (Encode::encode('shiftjis', "line ${line}: ${errmes}\n"));
}

sub checkNest
{
	my ($mode, $parent, $line, $param) = @_;
	my $errmes = '';

	if ($mode eq $TS_END && $parent ne $TS_BEGIN)
	{
		$errmes = "'\@$mode' は '$TS_BEGIN' と対になります。'$parent' とは対になれません。";
	}

	if (
		(
			$mode eq $TS_END ||
			$mode eq $TS_CONSTRUCTOR ||
			$mode eq $TS_METHOD
		) && (
			$parent ne $TS_BEGIN
		)
	)
	{
		$errmes = "'\@$mode' の親は '$TS_BEGIN' です。現在の親は '$parent' です。";
	}

	if (
		(	$mode eq $TS_GETTER ||
			$mode eq $TS_SETTER
		) && (
			$parent ne $TS_PROPERTY
		)
	)
	{
		$errmes = "'\@$mode' の親は '$TS_PROPERTY' です。現在の親は '$parent' です。";
	}

	if (
		(	$mode eq $TS_PARAMETER ||
			$mode eq $TS_PARAMS
		) && (
			$parent ne $TS_CONSTRUCTOR &&
			$parent ne $TS_METHOD &&
			$parent ne $TS_SETTER &&
			$parent ne $TS_FUNCTION
		)
	)
	{
		$errmes = "'\@$mode' の親は '$TS_CONSTRUCTOR' '$TS_METHOD' '$TS_FUNCTION' '$TS_SETTER' のいずれかです。現在の親は '$parent' です。";
	}

	if (
		(	$mode eq $TS_RETURN
		) && (
			$parent ne $TS_CONSTRUCTOR &&
			$parent ne $TS_METHOD &&
			$parent ne $TS_GETTER &&
			$parent ne $TS_FUNCTION
		)
	)
	{
		$errmes = "'\@$mode' の親は '$TS_CONSTRUCTOR' '$TS_METHOD' '$TS_FUNCTION' '$TS_GETTER' のいずれかです。現在の親は '$parent' です。";
	}


	if (
		(	$mode eq $TS_BEGIN ||
			$mode eq $TS_FUNCTION
		) && (
			$parent
		)
	)
	{
		$errmes = "'\@$mode' はトップレベルに記述します。";
	}

	if ($errmes) {
		abort($line, $errmes);
	}
}



# 
sub getInstanceCreator
{
	my $c = shift;
	my $rv;

	# プラグイン
	if ($PluginMode) {
		$rv = << "END";
// インスタンスクリエイター
static iTJSNativeInstance * Create_${TJSNI_PREFIX}${c}()
{
	return new ${TJSNI_PREFIX}${c}();
}
END
	}

	return $rv;
}


# クラスIDの定義。ネイティブでは何もしない。
sub getDefClassID
{
	my $c = shift;
	my $rv;

	# プラグイン
	if ($PluginMode) {
		my $decl = getDeclareClassID($c);
		$rv = << "END";
$TS_SEP
/*
 * TJS Native Class '${c}'
 * 
 * Native Instance Class   : ${TJSNI_PREFIX}${c}
 * Native Instance Creator : Create_${TJSNI_PREFIX}${c}()
 * Native Class Creator    : Create_${TJSNC_PREFIX}${c}()
 * Generated by $0
 */
#define TJS_NATIVE_CLASSID_NAME ClassID_${TJSNI_PREFIX}${c}
${decl}
END

	}
	return $rv;
}

# クラスIDの宣言。ネイティブでは何もしない。
sub getDeclareClassID
{
	my $c = shift;
	my $rv;

	# プラグイン
	if ($PluginMode) {
		unless (exists($UsingClasses{$c})) {
			$rv = "static tjs_int32 ClassID_${TJSNI_PREFIX}${c} = -1;";
			$UsingClasses{$c} = 1;
		}
	}
	return $rv;
}



# Class creator
sub getClassCreatorBegin
{
	my $c = shift;
	my $rv;

	# プラグイン
	if ($PluginMode) {
		$rv = << "END";
// クラスクリエイター
iTJSDispatch2 * Create_${TJSNC_PREFIX}${c}()
{
tTJSNativeClassForPlugin * classobj =
	TJSCreateNativeClassForPlugin(TJS_W("${c}"), Create_${TJSNI_PREFIX}${c});

// ネイティブメンバ定義部ここから
TJS_BEGIN_NATIVE_MEMBERS(${c})
${CurrentPP}
END

	# ネイティブ
	} else {
		my $ci = ($NoInstance) ? "return NULL;" : "return new ${TJSNI_PREFIX}${c}();";
		$rv = << "END";
//
// begin TJS Native Class '${c}'
//   Native Instance Class: ${TJSNI_PREFIX}${c}
//
// ネイティブクラス ${c} コンストラクタ
tTJSNativeInstance * ${TJSNC_PREFIX}${c}::CreateNativeInstance()
{
	${ci}
}

// クラスID
tjs_uint32 ${TJSNC_PREFIX}${c}::ClassID = (tjs_uint32)-1;

// ネイティブクラス ${c} のコンストラクタの実装
${TJSNC_PREFIX}${c}::${TJSNC_PREFIX}${c}() : tTJSNativeClass(TJS_W("${c}")) \
{ \
TJS_BEGIN_NATIVE_MEMBERS(${c}) \
${CurrentPP}
END
	}

	return $rv;
}


sub getClassCreatorEnd
{
	my $c = shift;
	my $rv;

	# プラグイン
	if ($PluginMode) {
		$rv = << "END";
// ネイティブメンバ定義部ここまで
TJS_END_NATIVE_MEMBERS

	/* この関数は classobj を返します。 */
	return classobj;
}

// 念のため
#undef TJS_NATIVE_CLASSID_NAME
// end of TJS Native Class '${c}'

END

	# ネイティブ
	} else {
		$rv = << "END";
	TJS_END_NATIVE_MEMBERS \
}
// end of TJS Native Class '${c}'
END
	}

	return $rv;
}

# Empty Finalize
sub getEmptyFinalizeDecl
{
	my $c = shift;
	my $rv = << "END";
// empty finalize method
TJS_DECL_EMPTY_FINALIZE_METHOD
END
	return $rv;
}

# Empty Constructor
sub getEmptyConstructorDecl
{
	my $c = shift;
	my $rv = getConstructorDeclBegin($c) . "\treturn TJS_S_OK;\n" . getConstructorDeclEnd($c);
	return $rv;
}


# Constructor
sub getConstructorDeclBegin
{
	my $c = shift;
	my $niflag = $NoInstance ? '_NO_INSTANCE' : '';
	my $macroarg;
	my $rv;

	if ($niflag) {
		$macroarg = "${c}";
	} else {
		$macroarg = << "END";
	_this, ${TJSNI_PREFIX}${c}, ${c}
END
	}

	$rv = << "END";
// constructor ${c}()
TJS_BEGIN_NATIVE_CONSTRUCTOR_DECL${niflag}(${macroarg})
{
	$TS_CONST_MEMBER_NAME_DECL = TJS_W("${c}.${c}");
	void * membername = NULL;	// DUMMY
${CurrentPP}
END

	return $rv;
}

sub getConstructorDeclEnd
{
	my $c = shift;
	my $stflag = $StaticDecl ? 'STATIC_' : '';

	my $rv = << "END";
}
TJS_END_NATIVE_${stflag}CONSTRUCTOR_DECL(${c})
// end of constructor ${c}()
END
	return $rv;
}


# Method
sub getMethodDeclBegin
{
	my $c = shift;
	my $m = shift;
	my $getni = getTJSGetNativeInstanceString($c);

	my $rv = << "END";
// method ${c}.${m}()
TJS_BEGIN_NATIVE_METHOD_DECL(${m})
{
	$TS_CONST_MEMBER_NAME_DECL = TJS_W("${c}.${m}");
	void * membername = NULL;	// DUMMY
	${getni}
${CurrentPP}
END
	return $rv;
}

sub getMethodDeclEnd
{
	my $c = shift;
	my $m = shift;
	my $stflag = $StaticDecl ? 'STATIC_' : '';

	my $rv = << "END";
}
TJS_END_NATIVE_${stflag}METHOD_DECL(${m})
// end of method ${c}.${m}()
END
	return $rv;
}


# Property
sub getPropDeclBegin
{
	my $c = shift;
	my $m = shift;
	my $stflag = $StaticDecl ? 'static ' : '';

	my $rv = << "END";
// ${stflag}property ${c}.${m}
TJS_BEGIN_NATIVE_PROP_DECL(${m})
{
END
	return $rv;
}

sub getPropDeclEnd
{
	my $c = shift;
	my $m = shift;
	my $stflag = $StaticDecl ? 'STATIC_' : '';

	my $rv = << "END";
}
TJS_END_NATIVE_${stflag}PROP_DECL(${m})
// end of property ${c}.${m}
END
	return $rv;
}




# Property getter
sub getGetterDeclBegin
{
	my $c = shift;
	my $m = shift;
	my $getni = getTJSGetNativeInstanceString($c);

	my $rv = << "END";
$TS_SEP
// getter ${c}.${m}
TJS_BEGIN_NATIVE_PROP_GETTER
{
	$TS_CONST_MEMBER_NAME_DECL = TJS_W("${c}.${m}");
	${getni}
${CurrentPP}
END
	return $rv;
}

sub getGetterDeclEnd
{
	my $c = shift;
	my $m = shift;
	my $rv = << "END";
}
TJS_END_NATIVE_PROP_GETTER
END
	return $rv;
}


# Setter
sub getSetterDeclBegin
{
	my $c = shift;
	my $m = shift;
	my $getni = getTJSGetNativeInstanceString($c);

	my $rv = << "END";
$TS_SEP
// setter ${c}.${m}
TJS_BEGIN_NATIVE_PROP_SETTER
{
	$TS_CONST_MEMBER_NAME_DECL = TJS_W("${c}.${m}");
	${getni}
	ZrvlParameterIterator curr_param(const_cast<tTJSVariant **>(&param), 1, 1, ${TS_MEMBER_NAME});
${CurrentPP}
END
	return $rv;
}

sub getSetterDeclEnd
{
	my $c = shift;
	my $m = shift;
	my $rv = << "END";
}
TJS_END_NATIVE_PROP_SETTER
${CurrentPP}
END
	return $rv;
}


# Getter (deny)
sub getDenyGetterDecl
{
	my $c = shift;
	my $m = shift;
	my $rv = << "END";
$TS_SEP
// getter ${c}.${m} is denied
TJS_DENY_NATIVE_PROP_GETTER
${CurrentPP}
END
	return $rv;
}


# Setter (deny)
sub getDenySetterDecl
{
	my $c = shift;
	my $m = shift;
	my $rv = << "END";
$TS_SEP
// setter ${c}.${m} is denied
TJS_DENY_NATIVE_PROP_SETTER
${CurrentPP}
END
	return $rv;
}


# Property (readonly)
sub getPropDeclReadOnly
{
	my $c = shift;
	my $m = shift;
	my $type  = shift;
	my $value = shift;

	my $stflag = $StaticDecl ? 'static ' : '';

	my $rv = getPropDeclBegin($c, $m);
	$rv .= getGetterDeclBegin($c, $m);
	my $ret = getReturnStatement($type, $value);

	# プラグイン
	if ($PluginMode) {
		$rv .= << "END";
$ret
	return TJS_S_OK;
${CurrentPP}
END

	# ネイティブ
	} else {
		$rv .= << "END";
	ZRVL_SET_RETURN_VALUE_${type}(${value})
	return TJS_S_OK;
${CurrentPP}
END

	}

	$rv .= getGetterDeclEnd($c, $m);
	$rv .= getDenySetterDecl($c, $m);
	$rv .= getPropDeclEnd($c, $m);

	return $rv;
}


sub getTJSGetNativeInstanceString
{
	my ($c) = @_;
	my $getni = ($NoInstance || $StaticDecl) ? 
		('// This is static member.') : 
		"TJS_GET_NATIVE_INSTANCE(_this, ${TJSNI_PREFIX}${c});";
	return $getni;
}

# 関数 BEGIN
sub getFunctionDeclBegin
{
	my $c = shift;

	my $cppclass = getCppClassName($c);
	my $rv;

	# プラグイン
	if ($PluginMode) {
		$rv = << "END";
// function ${c}()
tjs_error TJS_INTF_METHOD ${cppclass}::FuncCall(
	tjs_uint32 flag, const tjs_char * membername, tjs_uint32 *hint,
	tTJSVariant *result, tjs_int numparams, tTJSVariant **param, iTJSDispatch2 *objthis)
{
	$TS_CONST_MEMBER_NAME_DECL = TJS_W("${c}");
${CurrentPP}
END

	# ネイティブ
	} else {
		$rv = << "END";
// function ${c}()
class tjse_NativeFunctionClass_${c} : public tTJSNativeFunction
{
public:
	tjs_error Process(tTJSVariant *result, tjs_int numparams,
		tTJSVariant **param, iTJSDispatch2 *objthis)
	{
		$TS_CONST_MEMBER_NAME_DECL = TJS_W("${c}");
		void * membername = NULL;	// DUMMY
${CurrentPP}
END
	}

	return $rv;
}


# 関数 END
sub getFunctionDeclEnd
{
	my ($c) = @_;
	my $rv;

	# プラグイン
	if ($PluginMode) {
		$rv = << "END";
}
// end of function ${c}()
${CurrentPP}
END

	# ネイティブ
	} else {
		$rv = << "END";
	} // end of method
}; // end of class
{
	iTJSDispatch2 * global = ZrvlGetGlobalObject();
	iTJSDispatch2 * func = new tjse_NativeFunctionClass_${c}();
	tTJSVariant func_var(func); \
	func->Release(); \
	TJS_THROW_IF_ERROR(global->PropSet(TJS_MEMBERENSURE, TJS_W("${c}"), NULL, &func_var, NULL)); \
}
END
	}

	return $rv;
}

# パラメータの数（method, function用）
sub getNumParamsChecker
{
	my ($num) = @_;
	my $cmt = $num ? "パラメータは最低${num}個必要" : "パラメータ不要";

	my $rv = << "END";
	// ${cmt}
	ZrvlParameterIterator curr_param(param, numparams, ${num}, ${TS_MEMBER_NAME});
	if (membername) return TJS_E_MEMBERNOTFOUND;
	if (!curr_param.IsSufficient()) return TJS_E_BADPARAMCOUNT;
${CurrentPP}
END

	return $rv;
}

# パラメータの数（setter用）
sub getNumParamsCheckerForSetter
{
	my ($num) = @_;
	my $rv = "";
	return $rv;
}



# パラメータの受け取り
sub getParameterGetter
{
	my ($type, $varname, $default) = @_;
	my $decl = qq|#error 記述ミス: '\@${TS_PARAMETER} $type $varname $default'|;

	my $is_valid = "curr_param.IsValid()";
	my $reject_void = "curr_param.RejectVoid()";


	my $rv;

	# デフォルト値が指定されている場合
	if ($default ne '') {
		my %opt_decl_hash = (
'int'   => "tjs_int32  ${varname} = curr_param.AsInteger(${default}); curr_param.Next();",
'uint'  => "tjs_uint32 ${varname} = curr_param.AsUnsignedInteger(${default}); curr_param.Next();",
'int64' => "tjs_int64  ${varname} = curr_param.AsInteger64(${default}); curr_param.Next();",
'real'  => "tjs_real   ${varname} = curr_param.AsReal(${default}); curr_param.Next();",
'str'   => "ttstr      ${varname} = curr_param.AsString(${default}); curr_param.Next();",
'obj'      => "iTJSDispatch2  * ${varname} = curr_param.AsObjectNoAddRef(${default}); curr_param.Next();",
'obj+'     => "iTJSDispatch2  * ${varname} = curr_param.AsObject(${default}); curr_param.Next();",
'objthis'  => "iTJSDispatch2  * ${varname} = curr_param.AsObjectThisNoAddRef(${default}); curr_param.Next();",
'objthis+' => "iTJSDispatch2  * ${varname} = curr_param.AsObjectThis(${default}); curr_param.Next();",
'vstr' => "tTJSVariantString  * ${varname} = curr_param.AsVariantStringNoAddRef(${default}); curr_param.Next();",
'vstr+'=> "tTJSVariantString  * ${varname} = curr_param.AsVariantString(${default}); curr_param.Next();",
'voct' => "tTJSVariantOctet   * ${varname} = curr_param.AsVariantOctetNoAddRef(${default}); curr_param.Next();",
'voct+'=> "tTJSVariantOctet   * ${varname} = curr_param.AsVariantOctet(${default}); curr_param.Next();",
'vclo' => "tTJSVariantClosure * ${varname} = curr_param.AsObjectClosureNoAddRef(${default}); curr_param.Next();",
'vclo+'=> "tTJSVariantClosure * ${varname} = curr_param.AsObjectClosure(${default}); curr_param.Next();",
'var'  => "tTJSVariant * ${varname} = curr_param.AsVariant(${default}); curr_param.Next();",
		);
		if ($opt_decl_hash{$type}) {
			$decl = "$opt_decl_hash{$type}";
		}

		$rv = << "END";
	${decl}
${CurrentPP}
END

	# デフォルト値が指定されていない場合
	} else {
		my %req_decl_hash = (
'int'   => "tjs_int32  ${varname} = curr_param.AsInteger(); curr_param.Next();",
'uint'  => "tjs_uint32 ${varname} = curr_param.AsUnsignedInteger(); curr_param.Next();",
'int64' => "tjs_int64  ${varname} = curr_param.AsInteger64(); curr_param.Next();",
'real'  => "tjs_real   ${varname} = curr_param.AsReal(); curr_param.Next();",
'str'   => "ttstr      ${varname} = curr_param.AsString(); curr_param.Next();",
'obj'      => "iTJSDispatch2  * ${varname} = curr_param.AsObjectNoAddRef(); curr_param.Next();",
'obj+'     => "iTJSDispatch2  * ${varname} = curr_param.AsObject(); curr_param.Next();",
'objthis'  => "iTJSDispatch2  * ${varname} = curr_param.AsObjectThisNoAddRef(); curr_param.Next();",
'objthis+' => "iTJSDispatch2  * ${varname} = curr_param.AsObjectThis(); curr_param.Next();",
'vstr' => "tTJSVariantString  * ${varname} = curr_param.AsVariantStringNoAddRef(); curr_param.Next();",
'vstr+'=> "tTJSVariantString  * ${varname} = curr_param.AsVariantString(); curr_param.Next();",
'voct' => "tTJSVariantOctet   * ${varname} = curr_param.AsVariantOctetNoAddRef(); curr_param.Next();",
'voct+'=> "tTJSVariantOctet   * ${varname} = curr_param.AsVariantOctet(); curr_param.Next();",
'vclo' => "tTJSVariantClosure * ${varname} = curr_param.AsObjectClosureNoAddRef(); curr_param.Next();",
'vclo+'=> "tTJSVariantClosure * ${varname} = curr_param.AsObjectClosure(); curr_param.Next();",
'var'  => "tTJSVariant * ${varname} = curr_param.AsVariant(); curr_param.Next();",
		);
		if ($req_decl_hash{$type}) {
			$decl = $req_decl_hash{$type};
		}

		$rv = << "END";
	${reject_void};
	${decl}
${CurrentPP}
END
	}

	return $rv;
}


# 戻り値を設定
sub getReturnStatement
{
	my ($type, $value) = @_;
	my $decl = qq|#error 記述ミス '\@${TS_RETURN} $type $value'|;

	my %ret_hash = (
'int'  => "if (result) *result = (tTVInteger)(${value});",
'uint' => "if (result) *result = (tTVInteger)(${value});",
'real' => "if (result) *result = (tTVReal)(${value});",
'str'  => "if (result) *result = ${value};",
'var'  => "if (result) *result = ${value};",
'obj'  => "if (result) *result = tTJSVariant(${value}, ${value});",
'void' => "if (result) result->Clear();",
	);
	if ($ret_hash{$type}) {
		$decl = $ret_hash{$type};
	}

	my $rv = "\t" . $decl;
	return $rv;
}

# パラメータの型を制限
sub getParameterMust
{
	my ($type, $value) = @_;
	my $decl = qq|#error 記述ミス '\@${TS_PARAMMUST} $type $value'|;

	my %ret_hash = (
'str'        => "curr_param.IsStringOrThrow();",
'oct'        => "curr_param.IsOctetOrThrow();",
'obj'        => "curr_param.IsObjectOrThrow();",
'notnull'    => "curr_param.IsNotNullOrThrow();",
'notvoid'    => "curr_param.IsNotVoidOrThrow();",
'instanceof' => qq!curr_param.IsInstanceOfOrThrow(TJS_W("${value}"));!,
	);

	if ($ret_hash{$type}) {
		$decl = $ret_hash{$type};
	}

	my $rv = << "END";
	${decl}
${CurrentPP}
END
	return $rv;
}



