#!/usr/bin/perl
# オブジェクトインターフェース
use strict;

use utf8;
my $ENCODE_NAME = ':encoding(shift_jis)';
binmode(STDIN,  $ENCODE_NAME);
binmode(STDOUT, $ENCODE_NAME);

exit if ($#ARGV < 1);
my ($OutSrcH, $OutSrcC) = @ARGV;

my $HeaderID = '_TVP_PLUGIN_TJSDISPATCH2_INTERFACE_EX_AUTO_H_';

my $PropGets = [];
my $PropGetByNums = [];


my $Delim = '//' . ('-' x 78);



# PropGet
my $PropGetTemplate = << 'EOH';
###DELIMITER###
// property get ###CLASS_NAME###.###PROP_NAME###
/*TP*/###PROP_TYPE### TPPropGet_###CLASS_NAME###_###PROP_NAME###(iTJSDispatch2 * obj, const tjs_char * caller, bool ignoreProp)
{
	static tTJSVariant workVal;
	static tjs_uint32 hint = 0;
	tjs_uint32 flag = (ignoreProp) ? (TJS_MEMBERMUSTEXIST | TJS_IGNOREPROP) : TJS_MEMBERMUSTEXIST;
	if (TJS_FAILED(obj->PropGet(flag, TJS_W("###PROP_NAME###"), &hint, &workVal, obj))) {
		TJSE_THROW(TJSEM_E_PROPGET_FAIL, caller, TJS_W("###CLASS_NAME###.###PROP_NAME###"));
	}
	return ###TYPE_CAST###workVal###TYPE_CAST2###;
}

###DELIMITER###
// property get ###CLASS_NAME###.###PROP_NAME### (anonymous)
/*TP*/###PROP_TYPE### TPPropGet_###CLASS_NAME###_###PROP_NAME###(iTJSDispatch2 * obj, bool ignoreProp)
{
	return TPPropGet_###CLASS_NAME###_###PROP_NAME###(obj, TJSEM_ANON_FUNC, ignoreProp);
}

EOH

# PropGetByNum
my $PropGetByNumTemplate = << 'EOH';
###DELIMITER###
// property get (by num) ###CLASS_NAME###[n] as ###PROP_TYPE###
/*TP*/###PROP_TYPE### TPPropGetByNum_###CLASS_NAME###_As###TYPE_ID###(iTJSDispatch2 * obj, tjs_int num, const tjs_char * caller, bool ignoreProp)
{
	static tTJSVariant workVal;
	tjs_uint32 flag = (ignoreProp) ? (TJS_MEMBERMUSTEXIST | TJS_IGNOREPROP) : TJS_MEMBERMUSTEXIST;
	if (TJS_FAILED(obj->PropGetByNum(flag, num, &workVal, obj))) {
		TJSE_THROW(TJSEM_E_PROPGETBYNUM_FAIL, caller, TJS_W("###CLASS_NAME###"), (ttstr)num);
	}
	return ###TYPE_CAST###workVal###TYPE_CAST2###;
}

###DELIMITER###
// property get (by num) ###CLASS_NAME###[n] as ###PROP_TYPE###
/*TP*/###PROP_TYPE### TPPropGetByNum_###CLASS_NAME###_As###TYPE_ID###(iTJSDispatch2 * obj, tjs_int num, bool ignoreProp)
{
	return TPPropGetByNum_###CLASS_NAME###_As###TYPE_ID###(obj, num, TJSEM_ANON_FUNC, ignoreProp);
}
EOH


my $HeadSrc = << "END";
//----------------------------
// Generated by $0
//----------------------------

END


my $OutC = << "EOH";
$HeadSrc

#include "TVPPlugIn.h"
#include "$OutSrcH"

EOH

my $OutH = << "EOH";
$HeadSrc
#ifndef $HeaderID
#define $HeaderID

#include <windows.h>
#include "tp_stub.h"

EOH



while (<STDIN>) {
	chomp;
	$_  =~ s/^\s+//;
	next if (!$_ || $_ =~ /^#/);
	my ($a, $f) = split(/\s*,\s*/, $_, 2);
	if ($a && $f) {
		if ($a =~ s/^!//) {
			my ($class, $asname) = split(/\s+/, $f);
			push @$PropGetByNums, [$a, $class, $asname];
		} else {
			my ($class, $asname) = split(/\./, $f);
			push @$PropGets, [$a, $class, $asname];
		}

	}
}


foreach my $ea (@$PropGets) {
	my ($type, $class, $name) = @$ea;
	my $t = $PropGetTemplate;

	my ($cast, $cast2, $const) = getCast($type);
	my $op = '';

	$t =~ s/###PROP_TYPE###/$type/g;
	$t =~ s/###CLASS_NAME###/$class/g;
	$t =~ s/###PROP_NAME###/$name/g;
	$t =~ s/###TYPE_CAST###/$cast/g;
	$t =~ s/###TYPE_CAST2###/$cast2/g;
	$t =~ s/###DELIMITER###/$Delim/g;

	while ($t =~ m|/\*TP\*/(.*?)\n|sg) {
		$OutH .= "$1;\n";
	}

	$t =~ s|/\*TP\*/||g;
	$OutC .= "$t\n";
}

foreach my $ea (@$PropGetByNums) {
	my ($type, $class, $id) = @$ea;
	my $t = $PropGetByNumTemplate;
	my ($cast, $cast2, $const) = getCast($type);

	# reference
	if ($type =~ /.+\&$/) { $cast = "($type)" };

	$t =~ s/###PROP_TYPE###/$type/g;
	$t =~ s/###CLASS_NAME###/$class/g;
	$t =~ s/###TYPE_ID###/$id/g;
	$t =~ s/###TYPE_CAST###/$cast/g;
	$t =~ s/###TYPE_CAST2###/$cast2/g;
	$t =~ s/###DELIMITER###/$Delim/g;

	while ($t =~ m|/\*TP\*/(.*?)\n|sg) {
		$OutH .= "$1;\n";
	}

	$t =~ s|/\*TP\*/||g;
	$OutC .= "$t\n";
}

$OutH .= << "EOH";

#endif $HeaderID
EOH

open FC, ">$OutSrcC";
open FH, ">$OutSrcH";

binmode(FC,  $ENCODE_NAME);
binmode(FH,  $ENCODE_NAME);

print FC $OutC;
print FH $OutH;

close FC;
close FH;


sub getCast
{
	my ($type) = @_;

	my $cast = '';
	my $cast2 = '';
	my $const = $type =~ m/^const\s+/ ? 'const ' : '';

	# dispatch
	if ($type =~ m/iTJSDispatch2\s*\*$/) {
		$cast2 = ".AsObjectNoAddRef()";
	} elsif ($type =~ m/tTJSVariantOctet\s*\*$/) {
		$cast2 = ".AsOctetNoAddRef()";
	} elsif ($type =~ m/tTJSVariant$/) {
		$cast2 = "";
	} elsif ($type =~ m/tTJSVariantClosure\s*\*$/) {
		$cast  = "&(";
		$cast2 = ".AsObjectClosureNoAddRef())";
	} elsif ($type =~ m/.+\*$/) {
		$cast = "($type)(tTVInteger)";
	} elsif ($type eq 'tjs_real') {
		$cast = "(tTVReal)";
	} elsif ($type ne 'tjs_int') {
		$cast = "(tjs_int)";
	} elsif ($type ne 'tjs_int') {
		$cast = "($type)(tjs_int)";
	}

	return ($cast, $cast2, $const);
}

exit;
