#!/usr/bin/perl
# Makefileを作るためのスクリプト

use utf8;
my $ENCODE_NAME = ':encoding(shift_jis)';
binmode(STDIN,  $ENCODE_NAME);
binmode(STDOUT, $ENCODE_NAME);

my $SrcResFile = 'srcfiles.res';
my $SrcResDir = '..';

# 基本セット
my @Targets = qw(
generic
system
io
iconv
graph
sqlite
);
#もうちょっと…
#socket


# make追加分
my @ExtTargets = qw(
sqliteplus
sqlite-CLI
);

# ドキュメントが依存しているソース
my %DocDepends = (
'generic' => 'util_generic.ncc ArrayDic.ncc DateUtil.ncc RandomVariate.ncc Digest.ncc', 
);

# ソース追加分
my @ExtSrcTargets = qw(
sqliteplus
sqlite-CLI
);


# 依存
my %Depends = (
	'sqlite-CLI' => [ 'sqliteplus' ],
	'sqlite'     => [ 'common', 'sqliteplus' ],
	'sqliteplus' => [ 'common' ],
	'generic'    => [ 'common' ],
	'system'     => [ 'common' ],
	'socket'     => [ 'common' ],
	'graph'      => [ 'common' ],
	'io'         => [ 'common' ],
	'iconv'      => [ 'common' ],
	'common'     => [ ],
);


my $TargetList = join(' ', @Targets);
my @Out = ();

my $SEP = '=' x 32;


# --- MEMO
push @Out, << "END";
# -----------------------------
# Generated by $0
# -----------------------------

KRKR_PLUGIN_GLOBAL_CALL = 1

!if !$d(KRKR_PLUGIN_CONFIG_MAK)
!include Config.mak
!endif

END

# --- include
#push @Out, '!include Config.mak';
push @Out, '';

# --- dll
push @Out, "dll: implib common $TargetList " . join(' ', @ExtTargets);

# --- make
# Makefile自身がターゲット
push @Out, << 'END';
make:
	$(PERL)  tool/makemake.pl > Makefile
END

# --- make
# インポートライブラリ
push @Out, << 'END';
implib: ..\lib\iconv.lib ..\lib\sqlite3.lib
.dll.lib:
	@echo ================================================
	@echo $< -> $@
	@echo ================================================
	$(IMPORTER) -a $@ $<
END

# --- target
foreach my $e ('common', @ExtTargets, @Targets) {
	my $dep = $Depends{$e} ? join(' ', @{$Depends{$e}}) : '';
push @Out, << "END";
$e: $dep
	\@echo $SEP
	\@echo Making ${e} ...
	\@echo $SEP
	\@cd \$\@
	\$(MAKE)
	\@cd ..

END
}

# --- install
push @Out, makeRule('install', @Targets);
# --- clean
push @Out, makeRule('clean', 'common', @ExtTargets, @Targets) . "\n";
# --- src_clean
push @Out, makeRule('src_clean', 'common', @Targets);

# --- doc
{
my $docs;
my $docclean = "doc_clean:\n";
my $tgdoc = 'doc:';
foreach my $e (@Targets) {
	my $ef = "util_$e";
	my $nccs = $DocDepends{$e} || "${ef}.ncc";
	# ディレクトリを追記
	$nccs = join(' ', map { "$e/$_"; } (split(/\s+/, $nccs)));
	my $html = "\$(HTMLMAN_DIR)\\$ef\\index.html";
	my $tdf  = "\$(TDFMAN_DIR)\\$ef\\index.tdf";
	my $cmd  = qq{\$(TDF2HTML2) -s \$(SKELETON_FILE) -i \$(TDFMAN_DIR)/$ef -o \$(HTMLMAN_DIR)/$ef *.tdf};

	$tgdoc .= " $html";
$docs .= << "END";
$tdf: $nccs
	-mkdir \$(TDFMAN_DIR)\\$ef
	\$(PERL)  tool/makedoc.pl \$(TDFMAN_DIR)\\$ef index.tdf $nccs
$html: $tdf
	-mkdir \$(HTMLMAN_DIR)\\$ef
	$cmd

END

	# *.tdf は f_ で始まるものと index.tdf だけを消す
	$docclean .= << "END";
	-\$(RM) \$(TDFMAN_DIR)\\$ef\\f_*.tdf
	-\$(RM) \$(TDFMAN_DIR)\\$ef\\index.tdf
	-\$(RM) \$(HTMLMAN_DIR)\\$ef\\*.html
END

}

# 簡易メニュー

$tgdoc .= ' $(MAN_DIR)\index.html';
$docs .= << 'END';
$(MAN_DIR)\index.html: $(TDFMAN_DIR)\index.tdf
	$(TDF2HTML) -s $(SKELETON_FILE) $(TDFMAN_DIR)\index.tdf > $(MAN_DIR)\index.html

END

$docclean .= << "END";
	-\$(RM) \$(MAN_DIR)\\index.html
END

push @Out, $docs, $docclean, $tgdoc;
}


# --- release
# リリース用：べた書き ここから
{

# バイナリアーカイブ
my @RelFiles = qw(
CHANGES
readme.txt
lib/isqlite.exe
lib/sqlite3.dll
doc/index.html
doc/style
doc/images/*.png
doc/images/*.jpg
samples/hook_script
);

# ソースアーカイブ
my @SrcFiles = qw(
src/common/Makefile
src/common/readme.txt
src/common/kroi_props.txt
src/common/md5.h
src/common/md5.c
src/common/rmd_dgst.c
src/common/rmdconst.h
src/common/rmd_locl.h
src/common/ripemd.h
src/common/md32_common.h
src/common/OSVersionInfo.h
src/common/OSVersionInfo.cpp
src/common/GflSHA.h
src/common/GflSHA.cpp
src/common/TableHolder.h
src/common/TableHolder.cpp
src/common/TPMT19937ar.h
src/common/TPMT19937ar.cpp
src/common/TVPPlugIn.h
src/common/TVPPlugIn.cpp
src/common/TPPropGetIF.h
src/common/TPPropGetIF.h
src/common/TPException.h
src/common/TPException.cpp
src/common/ZrvlVariant.h
src/common/ZrvlVariant.cpp
src/common/ComplexRect.h
src/common/ComplexRect.cpp
src/common/tp_stub.h
src/common/tp_stub.cpp
src/common/TPRevision.h
src/common/sqlite3.h
src/common/TPMessage.h
src/common/TJSEMessageStr.h
src/common/tjse_ncdecl.h
src/common/maketab.c
src/tool/makehead.pl
src/tool/makekrnc.pl
src/tool/makekroi.pl
src/tool/makemain.pl
src/tool/makemake.pl
src/tool/makevsrc.pl
src/tool/update_revision_h.pl
src/tool/TVPPlugin.pm"
src/Makefile
src/Config.mak
src/Rules.mak
src/revision.mak
src/forms.txt
src/readme.txt
lib/dummy.txt
);


# デモアーカイブ（未使用）
my @DemoFiles = qw(
demo/image
demo/scenario/*.ks
demo/scenario/*.tjs
demo/system/Override.tjs
demo/startup.tjs
demo/readme.txt
);

# ソースファイル拡張子
my @RelSrcSuffix = qw(
h
c
cpp
ncc
tjs
txt
def
pl
);

# 各アーカイブ追加分
foreach my $e (@Targets) {
	my $dd = "src/${e}";
	push @RelFiles, "lib/util_${e}.dll";	# DLL
	push @RelFiles, "doc/html/util_${e}";	# Document HTML
	push @SrcFiles, "${dd}/Makefile";	# Makefile
	foreach my $e2 (@RelSrcSuffix) {
		push @SrcFiles, "${dd}/*.${e2}";
	}
}

# ソース追加分
foreach my $e (@ExtSrcTargets) {
	my $dd = "src/${e}";
	push @SrcFiles, "${dd}/Makefile";
	foreach my $e2 (@RelSrcSuffix) {
		push @SrcFiles, "${dd}/*.${e2}";
	}
}


# srcのみレスポンスファイル出力（数が多いので）
{
	my $srcfls  = join("\n", @SrcFiles);
	open (FS, ">${SrcResDir}/${SrcResFile}") or die("${SrcResDir}/${SrcResFile}: $!");
	binmode(FS, $ENCODE_NAME);
	print FS "$srcfls\n";
	close(FS);
}

my $relfls  = join(' ', @RelFiles);
my $demofls = join(' ', @DemoFiles);

push @Out, << 'END';
distclean: clean doc_clean
all: tdf_rev_info dll doc
release: distclean all src_clean release_only build_test

revision:
	$(PERL)  tool/update_revision_h.pl $(TP_VERSION) $(TVP_VERSION)

tdf_rev_info:
	$(PERL)  tool/maketdffile.pl $(TP_VERSION) $(TVP_VERSION)
END

push @Out, << "END";
release_only:
	\@cd ..
	-\$(RM) \$(TP_REL_FILE)
	-\$(RM) \$(TP_REL_SRC_FILE)
	-mzp -r \$(TP_REL_FILE) $relfls
	-mzp -r \$(TP_REL_SRC_FILE) \@${SrcResFile}
	\@cd src

END

# デモ
#	-\$(RM) \$(TP_REL_DEMO_FILE)
#	-mzp -r \$(TP_REL_DEMO_FILE) $demofls

}
# リリース用：べた書き ここまで



# --- svn tag
push @Out, << 'END';
SVN_REPOS  = http://cm6/svn/kr2
SVN_TRUNK  = $(SVN_REPOS)/trunk/tp_utils
SVN_TAGS   = $(SVN_REPOS)/tags/tp_utils
SVN_BRANCH = $(SVN_REPOS)/branch/$(SVN_BRANCH_NAME)/tp_utils
SVN_NEWTAG = tp_utils-$(TP_VERSION)-TVP$(TVP_VERSION)

# SVN: trunkからタグ付け
svn_tag_from_trunk:
	svn copy $(SVN_TRUNK) $(SVN_TAGS)/$(SVN_NEWTAG)

# SVN: branchからタグ付け
svn_tag_from_branch:
	svn copy $(SVN_BRANCH) $(SVN_TAGS)/$(SVN_NEWTAG)

END


# --- test build
push @Out, << 'END';

BUILD_TEST_DIR = ..\build_test

build_test:
	@echo ================================================
	@echo Build test ...
	@echo ================================================
	-rm -rf $(BUILD_TEST_DIR)/*
	-makedir -p $(BUILD_TEST_DIR)/lib
	-makedir -p $(BUILD_TEST_DIR)/src/common
	-cp ../$(TP_REL_SRC_FILE) $(BUILD_TEST_DIR)/
	-cp ../lib/iconv.dll $(BUILD_TEST_DIR)/lib/
	-cp ../lib/sqlite3.dll $(BUILD_TEST_DIR)/lib/
	-cp common/iconv.h $(BUILD_TEST_DIR)/src/common/
	-pushd $(BUILD_TEST_DIR)
	-cd $(BUILD_TEST_DIR)
	-jar xf src-$(TP_REL_FILE_BASE)
	-cd src
	make
	-popd
	@echo ================================================
	@echo Build test : done
	@echo ================================================
END



# 書き出し
print join("\n", @Out);

sub makeRule {
	my ($t, @target) = @_;
	my $rc = "$t:\n";
	foreach my $e (@target) {
		$rc .= << "END";
	cd $e
	\$(MAKE) $t
	\@cd ..
END
	}

	return $rc;
}

